<?php

require_once __DIR__ . '/vendor/autoload.php';
require_once __DIR__ . '/../FFmpeg.php';

use YouTube\YouTubeDownloader;

class YouTube extends Downloader
{
    private $hideDashVideos = false;
    private $showMp3 = false;
    private $enableRedirector = false;
    private $availableFormats = ['audio' => [], 'video' => [], 'dash' => []];

    public function fetch($videoUrl)
    {
        $this->configure();
        $clients = $this->getYouTubeClients();

        $yt = new YouTubeDownloader();
        if ($this->proxy !== null) {
            $yt->setProxy($this->proxy->getServer(), $this->proxy->getType(), $this->proxy->getCredentials());
        }

        $data = $yt->getDownloadLinks($videoUrl, ['clients' => $clients]);
        $this->processDownloadData($data);
    }

    private function configure()
    {
        $this->hideDashVideos = get_option('aiodl_hide_dash') == 'on';
        $this->showMp3 = get_option('aiodl_show_mp3') == 'on';
    }

    private function getYouTubeClients()
    {
        $http = new Http('https://api.aiodown.com/v1/youtube/clients');
        $http->run();
        return json_decode($http->response, true);
    }

    private function processDownloadData($data)
    {
        $formats = $data->getAllFormats();
        $info = $data->getInfo();

        if (!empty($formats) && !empty($info)) {
            $this->extractMediaData($formats, $info);
        }
    }

    private function extractMediaData($formats, $info)
    {
        $this->title = $info->title;
        $this->source = 'youtube';
        $this->thumbnail = 'https://i.ytimg.com/vi/' . $info->id . '/mqdefault.jpg';
        $this->duration = $info->durationSeconds;
        $defaultAudio = null;
        $copyrightProtected = $this->isCopyrightProtected($formats[0]->url);
        foreach ($formats as $media) {
            preg_match('/(audio|video)\/(.*?);/', $media->mimeType, $matches);
            $isAudio = $matches[1] == 'audio';
            $isDash = empty($media->audioQuality);
            if ($isDash && $this->hideDashVideos) {
                continue;
            }
            if (empty($media->contentLength)) {
                $http = new Http($media->url);
                $media->contentLength = $http->getFileSize();
            }
            if (empty($media->contentLength) || $media->contentLength <= 0) {
                continue;
            }
            if (!$isAudio && !empty($media->qualityLabel)) {
                $quality = $media->qualityLabel;
            } else if ($isAudio && !empty($media->audioQuality)) {
                $quality = $this->getAudioQuality($media->itag);
            } else {
                $quality = 'hd';
            }
            if ($quality === null) {
                continue;
            }
            $extension = $matches[2];
            if ($isAudio && $extension == 'mp4') {
                $extension = 'm4a';
            }
            if ($extension == '3gpp') {
                $extension = '3gp';
            }
            $stream = new Media($media->url, $quality, $extension, !$isAudio, !$isDash);
            $stream->url = $stream->url . '&title=' . urlencode($this->title);
            $stream->size = $media->contentLength;
            if ($isDash) {
                $this->addToAvailableList($stream, 'dash');
            } else if (!$isAudio) {
                $this->addToAvailableList($stream, 'video');
            } else {
                $this->addToAvailableList($stream, 'audio');
                if ($defaultAudio === null) {
                    $defaultAudio = $stream;
                }
            }
            if ($extension == 'm4a' && $this->showMp3) {
                $stream = new Media($media->url, $quality, 'mp3', !$isAudio, !$isDash);
                $stream->size = (int)$media->contentLength;
                $this->addToAvailableList($stream, 'audio');
            }
        }
        $ffmpegEnabled = get_option('aiodl_enable_ffmpeg') == 'on';
        if (!$copyrightProtected && $ffmpegEnabled) {
            $this->createRenderLinks($defaultAudio);
        }

        if ($this->enableRedirector) {
            $this->createRedirectedLinks();
        }
        $this->sortAndMergeMediaLists();
    }

    private function createRenderLinks($defaultAudio)
    {
        foreach ($this->availableFormats['dash'] as $media) {
            if ($media->quality !== '1080p' && $media->quality !== '720p' && $media->quality !== '360p') {
                continue;
            }
            if ($media->extension !== 'mp4') {
                continue;
            }
            if (!empty($this->availableFormats['video'][$media->quality . $media->extension])) {
                continue;
            }
            $ffmpegUrl = FFmpeg::combinedLinkSelfHosted($media->url, $defaultAudio->url, $this->title . '.mp4');
            $combined = new Media($ffmpegUrl, $media->quality, 'mp4', true, true);
            $combined->size = $media->size + $defaultAudio->size;
            $combined->requiresRendering = true;
            $this->addToAvailableList($combined, 'video');
        }
    }

    private function createRedirectedLinks()
    {
        foreach ($this->medias as $media) {
            $media->url = preg_replace(
                '/^(.*)\.googlevideo\.com/',
                'https://redirector.googlevideo.com',
                $media->url
            );
        }
    }

    private function getAudioQuality($itag)
    {
        $qualities = [
            '139' => '48kbps',
            '140' => '128kbps',
            '141' => '256kbps',
            '171' => '128kbps',
            '249' => '48kbps',
            '250' => '64kbps',
            '251' => '128kbps',
            '256' => '192kbps',
            '258' => '384kbps',
        ];
        if (!empty($qualities[$itag])) {
            return $qualities[$itag];
        }
        return null;
    }

    private function sortAndMergeMediaLists()
    {
        $this->sortMediaList('audio');
        $this->sortMediaList('video');
        $this->sortMediaList('dash');

        $this->medias = array_merge($this->medias, $this->availableFormats['video']);
        $this->medias = array_merge($this->medias, $this->availableFormats['audio']);
        $this->medias = array_merge($this->medias, $this->availableFormats['dash']);
    }

    private function sortMediaList($type)
    {
        usort($this->availableFormats[$type], ['Helpers', 'sortByQuality']);
    }

    private function isCopyrightProtected($url)
    {
        $query = parse_url($url, PHP_URL_QUERY);
        parse_str($query, $params);
        if (!empty($params['beids']) || !empty($params['gcr'])) {
            return true;
        }
        return false;
    }

    private function hasBetter($stream, $type)
    {
        $existingSize = -1;
        if (!empty($this->availableFormats[$type][$stream->quality . $stream->extension])) {
            $existingSize = $this->availableFormats[$type][$stream->quality . $stream->extension]->size;
        }
        return $existingSize > $stream->size;
    }

    private function addToAvailableList($stream, $type)
    {
        if (!$this->hasBetter($stream, $type)) {
            $this->availableFormats[$type][$stream->quality . $stream->extension] = $stream;
        }
    }

    public function getChunks($url, $size)
    {
        $chunks = [];
        $chunkSize = 10485760;
        $chunkCount = ceil($size / $chunkSize);
        for ($i = 0; $i < $chunkCount; $i++) {
            $start = $i * $chunkSize;
            $end = $start + $chunkSize - 1;
            if ($end >= $size) {
                $end = $size - 1;
            }
            $chunks[] = $url . '&range=' . $start . '-' . $end;
        }
        return $chunks;
    }
}